(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2017)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Ground Completion\<close>

theory Ground_Completion
  imports
    Abstract_Completion
    "../Normalization_Equivalence/Normalization_Equivalence"
    Completion_Fairness
begin

definition WCR1
  where
    "WCR1 r \<longleftrightarrow> r\<inverse> O r \<subseteq> r O r\<inverse> \<union> Id"

text \<open>Balanced weakly Church-Rosser\<close>
definition BWCR
  where
    "BWCR r \<longleftrightarrow> (\<forall>a b. (a, b) \<in> r\<inverse> O r \<longrightarrow> (\<exists>c n. (a, c) \<in> r ^^ n \<and> (b, c) \<in> r ^^ n))"

lemma BWCR_E:
  assumes "BWCR r"
    and "(a, b) \<in> r" and "(a, c) \<in> r"
  obtains d and n where "(b, d) \<in> r ^^ n" and "(c, d) \<in> r ^^ n"
  using assms unfolding BWCR_def by blast

lemma BWCR_imp_WCR: "BWCR r \<Longrightarrow> WCR r"
  by (auto simp: WCR_defs) (meson BWCR_E joinI relpow_imp_rtrancl)

lemma WCR1_imp_BWCR:
  assumes "WCR1 r"
  shows "BWCR r"
proof (unfold BWCR_def, intro allI impI)
  fix a b assume "(a, b) \<in> r\<inverse> O r"
  with assms consider "a = b" | c where "(a, c) \<in> r " and "(b, c) \<in> r" by (auto simp: WCR1_def)
  then show "\<exists>c n. (a, c) \<in> r ^^ n \<and> (b, c) \<in> r ^^ n"
  proof (cases)
    case 1
    then show ?thesis by (intro exI [of _ a] exI [of _ 0]) simp
  next
    case 2
    then show ?thesis by (intro exI [of _ c] exI [of _ 1]) simp
  qed
qed

lemma steps_CR_on:
  assumes "CR_on r {a}" and "(a, b) \<in> r\<^sup>*"
  shows "CR_on r {b}"
  using assms
  by (meson CR_onE CR_on_singletonI rtrancl_trans singletonI)

text \<open>Exercise 1.39(c)\<close>
lemma BWCR_NF:
  assumes "BWCR r" and "(a, b) \<in> r\<^sup>!"
  shows "CR_on r {a}" and "SN_on r {a}" and "\<And>m n. (a, b) \<in> r ^^ m \<Longrightarrow> (a, b) \<in> r ^^ n \<Longrightarrow> m = n"
proof -
  obtain n where "b \<in> NF r" and "(a, b) \<in> r ^^ n" using \<open>(a, b) \<in> r\<^sup>!\<close>
    by (meson normalizability_E rtrancl_len_E)
  { fix m n c assume "(a, b) \<in> r ^^ m" and "(a, c) \<in> r ^^ n"
    then have "n \<le> m \<and> (c, b) \<in> r ^^ (m - n)"
    proof (induct m arbitrary: a c n)
      case 0 then show ?case using \<open>b \<in> NF r\<close> by (auto elim: relpow_E2)
    next
      case IH: (Suc m)
      show ?case
      proof (cases n)
        case 0 then show ?thesis using IH by auto
      next
        case (Suc k)
        with \<open>(a, c) \<in> r ^^ n\<close> and \<open>(a, b) \<in> r ^^ Suc m\<close>
        obtain d and e where "(a, d) \<in> r" and "(a, e) \<in> r"
          and "(d, b) \<in> r ^^ m"
          and "(e, c) \<in> r ^^ k"
          by (meson relpow_Suc_E2)
        with \<open>BWCR r\<close> obtain l and f where "(d, f) \<in> r ^^ l" and "(e, f) \<in> r ^^ l"
          by (auto elim: BWCR_E)
        have "(d, b) \<in> r\<^sup>!" using \<open>b \<in> NF r\<close> and \<open>(d, b) \<in> r ^^ m\<close> by (auto dest: relpow_imp_rtrancl)
        from IH(1) [OF \<open>(d, b) \<in> r ^^ m\<close> \<open>(d, f) \<in> r ^^ l\<close>]
        have "l \<le> m" and "(f, b) \<in> r ^^ (m - l)" by auto
        with \<open>(e, f) \<in> r ^^ l\<close> have "(e, b) \<in> r ^^ m"
          using le_Suc_ex relpow_add by fastforce
        from IH(1) [OF \<open>(e, b) \<in> r ^^ m\<close> \<open>(e, c) \<in> r ^^ k\<close>]
        have "k \<le> m" and "(c, b) \<in> r ^^ (m - k)" by auto
        then show ?thesis unfolding Suc by simp
      qed
    qed }
  note * = this
  show "SN_on r {a}"
  proof
    fix f :: "nat \<Rightarrow> 'a" presume "f 0 = a" and **: "\<And>i. (f i, f (Suc i)) \<in> r"
    moreover then have "(a, f i) \<in> r ^^ i" for i by (induct i) auto
    ultimately have "f n = b" using * [OF \<open>(a, b) \<in> r ^^ n\<close> \<open>(a, f n) \<in> r ^^ n\<close>] by simp
    with \<open>b \<in> NF r\<close> and \<open>(f n, f (Suc n)) \<in> r\<close> show False by auto
  qed force+
  show "CR_on r {a}"
  proof
    fix c d assume "(a, c) \<in> r\<^sup>*" and "(a, d) \<in> r\<^sup>*"
    then obtain m and k where "(a, c) \<in> r ^^ m" and "(a, d) \<in> r ^^ k" by blast
    with * [OF \<open>(a, b) \<in> r ^^ n\<close>] have "(c, b) \<in> r ^^ (n - m)" and "(d, b) \<in> r ^^ (n - k)" by auto
    then have "(c, b) \<in> r\<^sup>*" and "(d, b) \<in> r\<^sup>*" by (auto simp: rtrancl_power)
    then show "(c, d) \<in> r\<^sup>\<down>" by blast
  qed
  show "(a, b) \<in> r ^^ k \<Longrightarrow> (a, b) \<in> r ^^ m \<Longrightarrow> k = m" for k m
    using * [of "max k m" b "min k m"] and \<open>b \<in> NF r\<close>
    by (cases "k \<le> m") (auto simp: max_def min_def elim: relpow_E2)
qed

locale ground_kb = kb_inf +
  assumes not_enc: "enc = False"
begin

lemma rstep_enc [simp]: "rstep_enc R = rstep R"
  unfolding rstep_enc_def by (simp add: not_enc)

inductive
  KBinv :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> bool" (infix "\<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse>" 55)
  where
    orientl: "s \<succ> t \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E - {(s, t)}, R \<union> {(s, t)})" |
    orientr: "t \<succ> s \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E - {(s, t)}, R \<union> {(t, s)})" |
    delete: "(s, s) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E - {(s, s)}, R)" |
    compose: "(t, u) \<in> rstep (R - {(s, t)}) \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E, (R - {(s, t)}) \<union> {(s, u)})" |
    simplifyl: "(s, u) \<in> rstep R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> ((E - {(s, t)}) \<union> {(u, t)}, R)" |
    simplifyr: "(t, u) \<in> rstep R \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> ((E - {(s, t)}) \<union> {(s, u)}, R)" |
    collapse: "(t, u) \<in> rstep_enc (R - {(t, s)}) \<Longrightarrow> (t, s) \<in> R \<Longrightarrow> (E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E \<union> {(u, s)}, R - {(t, s)})"

lemma KBinv_finite:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E', R')" and "finite (E \<union> R)"
  shows "finite (E' \<union> R')"
  using assms by (cases) auto

lemma KBinv_imp_KB:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E', R')"
  shows "(E, R) \<turnstile>\<^sub>K\<^sub>B (E', R')"
  using assms by (cases; blast intro: KB.intros)

definition MSET :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) term multiset"
  where
    "MSET R = \<Union>#{#{#s, t#}. (s, t) \<in># mset_set (fst R)#} + \<Union>#{#{#s, t#}. (s, t) \<in># mset_set (snd R)#}"

lemma MSET: "MSET (E, R) =
  \<Union>#{#{#s, t#}. (s, t) \<in># mset_set E#} + \<Union>#{#{#s, t#}. (s, t) \<in># mset_set R#}"
  by (auto simp: MSET_def)

fun PAIR :: "('a, 'b) trs \<times> ('a, 'b) trs \<Rightarrow> ('a, 'b) term multiset \<times> ('a, 'b) rule set"
  where
    "PAIR (E, R) = (MSET (E, R), E)"

abbreviation "GT \<equiv> lex_two (s_mul_ext Id {\<succ>}) Id ((measure card)\<inverse>)"

lemma E_subset_MSET: "finite E \<Longrightarrow> (s, t) \<in> E \<Longrightarrow> {#s, t#} \<subseteq># MSET (E, R)"
  apply (auto simp: MSET_def)
  by (metis (mono_tags, lifting) finite_set_mset_mset_set mset_add mset_subset_eq_add_left old.prod.case subset_mset.add_increasing2 subset_mset.le_add_same_cancel1 sum_mset.insert)

lemma R_subset_MSET: "finite R \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> {#s, t#} \<subseteq># MSET (E, R)"
  by (metis E_subset_MSET MSET add.commute)

lemma MSET_Diff_insert:
  assumes "finite E" and "finite R"
    and "(s, t) \<in> E"
  shows "MSET (E - {(s, t)}, insert (s, t) R) =
    (if (s, t) \<in> R then MSET (E, R) - {#s, t#} else MSET (E, R))"
  using assms
  apply (auto simp: MSET_def mset_set_Diff image_mset_Diff)
   apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set E#}")
    apply (auto simp: sum_mset.remove insert_absorb)
  apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set E#}")
   apply (auto simp: sum_mset.remove insert_absorb)
  done

lemma MSET_Diff_insert':
  assumes "finite E" and "finite R"
    and "(s, t) \<in> E"
  shows "MSET (E - {(s, t)}, insert (t, s) R) =
    (if (t, s) \<in> R then MSET (E, R) - {#s, t#} else MSET (E, R))"
  using assms
  apply (auto simp: MSET_def mset_set_Diff image_mset_Diff)
   apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set E#}")
    apply (auto simp: sum_mset.remove insert_absorb)
  apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set E#}")
   apply (auto simp: sum_mset.remove insert_absorb)
  done

lemma MSET_E_Diff:
  assumes "finite E" and "(s, t) \<in> E" shows "MSET (E - {(s, t)}, R) = MSET (E, R) - {#s, t#}"
  using assms
  apply (auto simp: MSET_def mset_set_Diff image_mset_Diff)
  apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set E#}")
   apply (auto simp: sum_mset.remove)
  done

lemma MSET_R_Diff:
  assumes "finite R" and "(s, t) \<in> R" shows "MSET (E, R - {(s, t)}) = MSET (E, R) - {#s, t#}"
  using assms
  apply (auto simp: MSET_def mset_set_Diff image_mset_Diff)
  apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set R#}")
   apply (auto simp: sum_mset.remove)
  done

lemma MSET_E_insert:
  assumes "finite E"
  shows "MSET (insert (s, t) E, R) = (if (s, t) \<in> E then MSET (E, R) else MSET (E, R) + {#s, t#})"
  using assms
  apply (auto simp: MSET_def mset_set.insert_remove mset_set_Diff image_mset_Diff)
  apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set E#}")
   apply (auto simp: sum_mset.remove)
  done

lemma MSET_R_insert:
  assumes "finite R"
  shows "MSET (E, insert (s, t) R) = (if (s, t) \<in> R then MSET (E, R) else MSET (E, R) + {#s, t#})"
  using assms
  apply (auto simp: MSET_def mset_set.insert_remove mset_set_Diff image_mset_Diff)
  apply (subgoal_tac "{#{#s, t#}#} \<subseteq># {#{#s, t#}. (s, t) \<in># mset_set R#}")
   apply (auto simp: sum_mset.remove)
  done

lemma R_MSET:
  assumes "finite R" and "(s, t) \<in> R"
  shows "s \<in># MSET (E, R) \<and> t \<in># MSET (E, R)"
  using assms by (force simp: MSET)

lemma E_MSET:
  assumes "finite E" and "(s, t) \<in> E"
  shows "s \<in># MSET (E, R) \<and> t \<in># MSET (E, R)"
  using assms by (force simp: MSET)

lemma add_mset_Diff2:
  "N = {#x, y#} \<Longrightarrow> {#x, y#} \<subseteq># M \<Longrightarrow> add_mset y (M - N) = M - {#x#}"
  by (auto simp: multiset_eq_iff subseteq_mset_def split: if_splits)

context
  fixes R
  assumes R_less: "R \<subseteq> {\<succ>}"
begin

lemma KBinv_less:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E', R')"
  shows "R' \<subseteq> {\<succ>}"
  using KBinv_imp_KB [OF assms] and R_less
  using kb.KB_rtrancl_rules_subset_less kb_axioms by blast

lemma rstep_less:
  assumes "S \<subseteq> R" and "(s, t) \<in> rstep S"
  shows "s \<succ> t"
  using assms
  using R_less compatible_rstep_imp_less by blast

lemma KBinv_in_GT:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E', R')" and "finite (E \<union> R)"
  shows "(PAIR (E, R), PAIR (E', R')) \<in> GT"
  using assms
  apply (cases)
  subgoal for s t
    apply (cases "(s, t) \<in> R")
     apply (simp_all add: MSET_Diff_insert)
     apply (subgoal_tac "MSET (E, R) - {#s, t#} \<subset># MSET (E, R)")
      apply (rule disjI1)
      apply (rule subset_mset_imp_s_mul_ext_Id)
      apply assumption
     apply (metis (no_types, lifting) E_subset_MSET add_diff_cancel_left' add_mset_remove_trivial cancel_comm_monoid_add_class.diff_cancel diff_subset_eq_self multi_self_add_other_not_self subset_mset.diff_add subset_mset_def zero_diff)
    by (metis card_gt_0_iff diff_Suc_less equals0D)
  subgoal for t s
     apply (cases "(t, s) \<in> R")
     apply (simp_all add: MSET_Diff_insert')
     apply (subgoal_tac "MSET (E, R) - {#s, t#} \<subset># MSET (E, R)")
      apply (rule disjI1)
      apply (rule subset_mset_imp_s_mul_ext_Id)
      apply assumption
     apply (metis (no_types, lifting) E_subset_MSET add_diff_cancel_left' add_mset_remove_trivial cancel_comm_monoid_add_class.diff_cancel diff_subset_eq_self multi_self_add_other_not_self subset_mset.diff_add subset_mset_def zero_diff)
    by (metis card_gt_0_iff diff_Suc_less equals0D)
  subgoal for s
    apply (simp add: MSET_E_Diff)
    apply (rule disjI1)
    apply (rule subset_mset_imp_s_mul_ext_Id)
    by (metis (no_types, lifting) E_subset_MSET add_diff_cancel_left' add_mset_not_empty cancel_comm_monoid_add_class.diff_cancel diff_subset_eq_self subset_mset.diff_add subset_mset.dual_order.not_eq_order_implies_strict)
  subgoal for t u s
    apply (subgoal_tac "u \<noteq> t")
     apply (cases "(s, u) \<in> R")
      apply (simp_all add: MSET_R_insert MSET_R_Diff)
      apply (rule subset_mset_imp_s_mul_ext_Id)
      apply (metis (no_types, lifting) R_subset_MSET add_diff_cancel_left' add_mset_not_empty cancel_comm_monoid_add_class.diff_cancel diff_subset_eq_self subset_mset.diff_add subset_mset.order.not_eq_order_implies_strict)
     apply (rule s_mul_ext_IdI [of "{#t#}" _ "MSET (E, R) - {#t#}" _ "{#u#}"])
    using rstep_less [of "R - {(s, t)}" t u, OF Diff_subset]
        apply (simp_all add: R_MSET less_neq)
    apply (subgoal_tac "s \<in># MSET (E, R)")
     apply (simp add: multiset_eq_iff)
    using irrefl rstep_less apply blast
    apply (simp add: R_MSET)
    done
  subgoal for s u t
    apply (subgoal_tac "u \<noteq> s")
     apply (cases "(u, t) \<in> E")
      apply (simp_all add: MSET_E_insert MSET_E_Diff)
      apply (rule disjI1)
      apply (rule subset_mset_imp_s_mul_ext_Id)
      apply (metis (no_types, lifting) E_subset_MSET add_diff_cancel_left' add_mset_not_empty cancel_comm_monoid_add_class.diff_cancel diff_subset_eq_self subset_mset.diff_add subset_mset.order.not_eq_order_implies_strict)
     apply (rule disjI1)
     apply (rule s_mul_ext_IdI [of "{#s#}" _ "MSET (E, R) - {#s#}" _ "{#u#}"])
    using rstep_less [of "R" s u, OF subset_refl]
        apply (simp_all add: E_MSET less_neq)
    apply (intro add_mset_Diff2 [OF refl])
    apply (simp add: E_subset_MSET)
    done
  subgoal for t u s
    apply (subgoal_tac "t \<noteq> u")
      apply (cases "(s, u) \<in> E")
      apply (simp_all add: MSET_E_insert MSET_E_Diff)
      apply (rule disjI1)
      apply (rule subset_mset_imp_s_mul_ext_Id)
      apply (metis (no_types, lifting) E_subset_MSET add_diff_cancel_left' add_mset_not_empty cancel_comm_monoid_add_class.diff_cancel diff_subset_eq_self subset_mset.diff_add subset_mset.order.not_eq_order_implies_strict)
           apply (rule disjI1)
     apply (rule s_mul_ext_IdI [of "{#t#}" _ "MSET (E, R) - {#t#}" _ "{#u#}"])
    using rstep_less [of "R" t u, OF subset_refl]
        apply (simp_all add: E_MSET less_neq)
    apply (intro add_mset_Diff2 [of "{#s, t#}"])
     apply (simp)
    apply (simp add: E_subset_MSET add_mset_commute)
    done
  subgoal for t u s
    apply (simp_all)
    apply (rule disjI1)
    apply (cases "(u, s) \<in> E")
      apply (simp add: insert_absorb)
     apply (rule s_mul_ext_IdI [of "{#t, s#}" _ "MSET (E, R) - {#t, s#}" _ "{#}"])
        apply (simp_all)
      apply (subgoal_tac "{#t, s#} \<subseteq># MSET (E, R)")
       apply (simp add: R_MSET add_mset_Diff2)
      apply (simp add: R_subset_MSET)
    using MSET_R_Diff apply blast
    apply (simp add: MSET_E_insert)
    apply (rule s_mul_ext_IdI [of "{#t#}" _ "MSET (E, R) - {#t#}" _ "{#u#}"])
       apply (simp_all add: R_MSET)
     apply (simp add: MSET_R_Diff R_subset_MSET add_mset_Diff2)
    by (meson Diff_subset kb.rstep_encD kb_axioms rstep_less)
  done

end

lemma SN_GT: "SN GT"
  apply (rule lex_two)
    apply simp
   apply (rule SN_s_mul_ext)
    apply (unfold_locales)
        apply (auto simp: refl_on_def trans_less SN_less trans_Id wf_imp_SN)
  done

lemmas SN_inv_image_GT_PAIR = SN_inv_image [OF SN_GT, of PAIR]

(*TODO: move*)
definition "ground_trs R \<longleftrightarrow> (\<forall>(l, r) \<in> R. ground l \<and> ground r)"

lemma ground_trs_empty [simp]: "ground_trs {}" by (simp add: ground_trs_def)

lemma in_ground_trs: "ground_trs R \<Longrightarrow> (s, t) \<in> R \<Longrightarrow> ground s \<and> ground t"
  by (auto simp: ground_trs_def)

lemma ground_trs_Diff [simp]:
  "ground_trs A \<Longrightarrow> ground_trs (A - B)"
  by (auto simp: ground_trs_def)

lemma ground_trs_insert [simp]:
  "ground_trs (insert (s, t) A) \<longleftrightarrow> ground s \<and> ground t \<and> ground_trs A"
  by (auto simp: ground_trs_def)

lemma ground_trs_rstep:
  assumes "ground_trs R" and "S \<subseteq> R" and "ground t" and "(t, u) \<in> rstep S"
  shows "ground u"
  using assms
    apply (auto elim!: rstepE dest!: in_ground_trs)
  by (metis ground_subst ground_subst_apply)

lemma KBinv_ground_trs:
  assumes "(E, R) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E', R')"
    and "ground_trs E" and "ground_trs R"
  shows "ground_trs E' \<and> ground_trs R'"
  using assms(1)
proof (cases)
  case (compose t u s)
  then show ?thesis
    using assms and ground_trs_rstep [of R "R - {(s, t)}" t u]
    by (simp_all add: in_ground_trs) blast
next
  case (simplifyl s u t)
  then show ?thesis
    using assms and ground_trs_rstep [of R "R" s u]
    by (simp_all add: in_ground_trs)
next
  case (simplifyr t u s)
  then show ?thesis
    using assms and ground_trs_rstep [of R "R" t u]
    by (simp_all add: in_ground_trs)
next
  case (collapse t u s)
  then show ?thesis
    using assms and ground_trs_rstep [of R "R - {(t, s)}" t u]
    by (simp_all add: in_ground_trs) (blast)
qed (insert assms, simp_all add: in_ground_trs)

abbreviation "KB_INV \<equiv> {(X, Y). X \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> Y}"

lemma SN_on_KB_INV:
  assumes "finite E\<^sub>0"
  shows "SN_on KB_INV {(E\<^sub>0, {})}"
proof
  fix f assume "f 0 \<in> {(E\<^sub>0, {})}" and "\<forall>i. (f i, f (Suc i)) \<in> KB_INV"
  then have f0: "f 0 = (E\<^sub>0, {})" and *: "\<And>i. (f i, f (Suc i)) \<in> KB_INV" by auto
  have "\<forall>i. f i = ((fst \<circ> f) i, (snd \<circ> f) i)" by auto
  then have "\<exists>E R. \<forall>i. f i = (E i, R i)" by blast
  then obtain E and R where [simp]: "f i = (E i, R i)" for i by blast
  have "finite (E i \<union> R i)" for i
    using f0 and * and assms by (induct i) (auto dest: KBinv_finite)
  moreover have "R i \<subseteq> {\<succ>}" for i
    using f0 and * by (induct i) (auto dest: KBinv_less [THEN subsetD])
  ultimately have "(PAIR (E i, R i), PAIR (E (Suc i), R (Suc i))) \<in> GT" for i
    using * by (intro KBinv_in_GT) simp_all
  then have "(PAIR (f i), PAIR (f (Suc i))) \<in> GT" for i by simp
  with SN_inv_image_GT_PAIR show False unfolding SN_defs inv_image_def by blast
qed

lemma min_elt_Ex:
  assumes "finite E"
  shows "\<exists>E' R'. ((E, ({}::('a, 'b) trs)), (E', R')) \<in> KB_INV\<^sup>* \<and> (E', R') \<in> NF KB_INV"
proof (induct rule: SN_on_induct' [OF SN_on_KB_INV [OF assms]])
  case IH: (1 ER)
  show ?case
  proof (cases "ER \<in> NF KB_INV")
    case True
    show ?thesis
      by (rule exI [of _ "fst ER"], rule exI [of _ "snd ER"]) (auto simp: True)
  next
    case False
    then obtain E' R' where *: "(ER, (E', R')) \<in> KB_INV" by (force simp: NF_def)
    moreover obtain E'' R'' where "((E', R'), (E'', R'')) \<in> KB_INV\<^sup>*"
      and "(E'', R'') \<in> NF KB_INV"  using IH [OF *] by blast
    ultimately show ?thesis
      by (meson converse_rtrancl_into_rtrancl)
  qed
qed

definition "max_run n E R \<longleftrightarrow> (\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E (Suc i), R (Suc i))) \<and> (E n, R n) \<in> NF KB_INV"

lemma max_run_Ex:
  assumes "finite E\<^sub>0"
  shows "\<exists>n E R. E 0 = E\<^sub>0 \<and> R 0 = {} \<and> max_run n E R"
proof -
  from min_elt_Ex [OF assms] obtain E' and R'
    where "((E\<^sub>0, {}), (E', R')) \<in> KB_INV\<^sup>*" and [simp]: "(E', R') \<in> NF KB_INV" by blast
  then obtain f and n where [simp]: "f 0 = (E\<^sub>0, {})" "f n = (E', R')"
    and *: "i < n \<Longrightarrow> (f i, f (Suc i)) \<in> KB_INV" for i unfolding rtrancl_fun_conv by auto
  show ?thesis
    by (rule exI [of _ n], rule exI [of _ "fst \<circ> f"], rule exI [of _ "snd \<circ> f"])
      (insert *, auto simp: max_run_def)
qed

lemma variant_in_ground_trs:
  assumes "ground_trs R" and "p \<bullet> r \<in> R"
  shows "p \<bullet> r = r"
proof -
  obtain u and v where [simp]: "r = (u, v)" by (cases r)
  have "p \<bullet> r = (p \<bullet> u, p \<bullet> v)" by (simp add: eqvt)
  with assms have "ground (p \<bullet> u)" and "ground (p \<bullet> v)" by (auto simp: ground_trs_def)
  then have "ground u" and "ground v" by (simp_all add: eqvt)
  then have "u \<cdot> sop p = u" and "v \<cdot> sop p = v"
    using ground_subst_apply by blast+
  moreover have "p \<bullet> r = (u \<cdot> sop p, v \<cdot> sop p)" by (simp add: eqvt)
  ultimately show ?thesis by simp
qed

lemma in_ground_trs_vars_rule:
  assumes "ground_trs R" and "r \<in> R"
  shows "vars_rule r = {}"
  using assms by (auto simp: vars_defs ground_trs_def ground_vars_term_empty)

lemma NF_KB_INV_PCP_empty:
  assumes "ground_trs R" and NF: "(E, R) \<in> NF KB_INV"
  shows "PCP R = {}"
proof (rule ccontr)
  assume "PCP R \<noteq> {}"
  then obtain s and t where "(s, t) \<in> PCP R" by auto
  then obtain l\<^sub>1 r\<^sub>1 and p and l\<^sub>2 r\<^sub>2 where o: "overlap R R (l\<^sub>1, r\<^sub>1) p (l\<^sub>2, r\<^sub>2)" by (auto simp: PCP_def)
  then have rules: "(l\<^sub>1, r\<^sub>1) \<in> R" "(l\<^sub>2, r\<^sub>2) \<in> R"
    and p: "p \<in> funposs l\<^sub>2"
    and *: "p = \<epsilon> \<longrightarrow> (l\<^sub>1, r\<^sub>1) \<noteq> (l\<^sub>2, r\<^sub>2)"
    using \<open>ground_trs R\<close>
    by (auto simp: overlap_def dest: variant_in_ground_trs)
  then have "ground l\<^sub>1" and "ground (l\<^sub>2 |_ p)"
    using p [THEN funposs_imp_poss] and \<open>ground_trs R\<close>
    by (auto simp: in_ground_trs) (metis ctxt_supt_id ground_ctxt_apply in_ground_trs)+
  with o have [simp]: "l\<^sub>1 = l\<^sub>2 |_ p" by (auto simp: overlap_def ground_subst_apply)
  have rule1: "(l\<^sub>1, r\<^sub>1) \<in> R - {(l\<^sub>2, r\<^sub>2)}" (is "_ \<in> ?R")
    using * and rules and p [THEN funposs_imp_poss] apply auto
    using subt_at_id_imp_eps by blast
  have "(l\<^sub>2, replace_at l\<^sub>2 p r\<^sub>1) \<in> rstep_enc (R - {(l\<^sub>2, r\<^sub>2)})" (is "(_, ?r) \<in> _")
    using p [THEN funposs_imp_poss]
      and rstepI [OF rule1, of l\<^sub>2 "ctxt_of_pos_term p l\<^sub>2" Var, OF _ refl]
    by (auto simp: ctxt_supt_id)
  from KBinv.collapse [OF this rules(2)] and NF show False by auto
qed

lemma ground_max_run_canonical:
  assumes max_run: "max_run n E R" and R0: "R 0 = {}"
    and ground: "ground_trs (E 0)"
    and total: "\<And>s t. s = t \<or> s \<succ> t \<or> t \<succ> s"
  shows "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>* \<and> canonical (R n)"
proof -
  have KBinv: "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B\<^sub>\<inverse> (E (Suc i), R (Suc i))"
    and NF: "(E n, R n) \<in> NF KB_INV"
    using max_run by (auto simp: max_run_def)
  then have KB: "\<forall>i<n. (E i, R i) \<turnstile>\<^sub>K\<^sub>B (E (Suc i), R (Suc i))" by (auto dest: KBinv_imp_KB)
  then have "KB\<^sup>*\<^sup>* (E 0, {}) (E n, R n)"
    unfolding rtrancl_fun_conv [to_pred]
    by (intro exI [of _ "\<lambda>i. (E i, R i)"] exI [of _ n]) (auto simp: R0)
  from KB_rtrancl_rules_subset_less [OF this]
  have [dest]: "s \<succ> t" if "(s, t) \<in> R n" for s t using that by auto
  have En: "E n = {}"
  proof (rule ccontr)
    assume "E n \<noteq> {}"
    then obtain s t where eq: "(s, t) \<in> E n" by auto
    consider "s = t" | "s \<succ> t" | "t \<succ> s" using total by blast
    then show False
    proof (cases)
      case 1
      from KBinv.delete [OF eq [unfolded this]] show ?thesis using NF by blast
    next
      case 2
      from KBinv.orientl [OF this eq] show ?thesis using NF by blast
    next
      case 3
      from KBinv.orientr [OF this eq] show ?thesis using NF by blast
    qed
  qed
  have "ground_trs (E i)" "ground_trs (R i)" if "i \<le> n" for i
    using ground and R0 and KBinv and that
    by (induct i) (auto dest: KBinv_ground_trs Suc_le_lessD)
  then have ground_Rn: "ground_trs (R n)" by simp
  have [simp]: "PCP (R n) = {}"
    using max_run by (auto simp: max_run_def dest: NF_KB_INV_PCP_empty [OF ground_Rn])

  from finite_fair_run [of R E, OF R0 En KB]
  have "(rstep (E 0))\<^sup>\<leftrightarrow>\<^sup>* = (rstep (R n))\<^sup>\<leftrightarrow>\<^sup>*"
    and "SN (rstep (R n))" and "CR (rstep (R n))" by auto
  moreover
  have "reduced (R n)"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then consider "\<not> left_reduced (R n)" | "\<not> right_reduced (R n)" by (auto simp: reduced_def)
    then show False
    proof (cases)
      case 1
      then obtain l r where lr: "(l, r) \<in> R n" and "l \<notin> NF (rstep (R n - {(l, r)}))"
        by (auto simp: left_reduced_def)
      then obtain u where "(l, u) \<in> rstep (R n - {(l, r)})" by auto
      from KBinv.collapse [unfolded rstep_enc, OF this lr] and NF show ?thesis by blast
    next
      case 2
      then obtain l r where lr: "(l, r) \<in> R n" and "r \<notin> NF (rstep (R n))"
        by (auto simp: right_reduced_def)
      then obtain u where *: "(r, u) \<in> rstep (R n)" by auto
      moreover have "(r, u) \<notin> rstep {(l, r)}"
      proof
        assume "(r, u) \<in> rstep {(l, r)}"
        then obtain C and \<sigma> where r: "r = C\<langle>l \<cdot> \<sigma>\<rangle>" and "u = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
        have "l \<succ> r" using lr by blast
        then have "r \<succ> C\<langle>r \<cdot> \<sigma>\<rangle>" by (subst r) (auto dest: subst ctxt)
        moreover define f where "f i = ((\<lambda>t. C\<langle>t \<cdot> \<sigma>\<rangle>) ^^ i) r" for i
        ultimately have "f i \<succ> f (Suc i)" for i
          using subst and ctxt by (induct i) (fastforce)+
        then show False using SN_less by (auto)
      qed
      ultimately have "(r, u) \<in> rstep (R n - {(l, r)})" by blast
      from KBinv.compose [OF this lr] and NF show ?thesis by blast
    qed
  qed
  ultimately show ?thesis by (auto simp: canonical_def)
qed

end

end
